package asset

import (
	"context"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"strings"
)

type FileStorage struct {
	Base string
}

func NewFileStorage(base string) Storage {
	return &FileStorage{
		Base: base,
	}
}

var _ Storage = FileStorage{}

func (f FileStorage) Close() error {
	return nil
}

func (f FileStorage) Create(ctx context.Context, fileName string, reader io.Reader) (string, error) {
	fullPath := filepath.Join(f.Base, fileName)

	exists, err := f.Exists(ctx, fileName)
	if err != nil {
		return "", err
	}
	if exists {
		return "", fmt.Errorf("file exists")
	}

	out, err := os.Create(fullPath)
	if err != nil {
		return "", err
	}
	defer out.Close()

	_, err = io.Copy(out, reader)
	if err != nil {
		return "", err
	}
	return fmt.Sprintf("file://%s", fullPath), err
}

func (f FileStorage) Delete(ctx context.Context, location string) error {
	if strings.HasPrefix(location, fmt.Sprintf("file://%s", f.Base)) {
		return fmt.Errorf("invalid file storage location: %s", location)
	}
	location = strings.TrimPrefix(location, "file://")

	return os.Remove(location)
}

func (f FileStorage) Exists(ctx context.Context, location string) (bool, error) {
	if strings.HasPrefix(location, fmt.Sprintf("file://%s", f.Base)) {
		return false, fmt.Errorf("invalid file storage location: %s", location)
	}
	location = strings.TrimPrefix(location, "file://")

	_, err := os.Stat(location)
	if err != nil {
		if os.IsNotExist(err) {
			return false, nil
		}
		return false, err
	}
	return true, nil
}

func (f FileStorage) Read(ctx context.Context, location string) (io.ReadCloser, error) {
	if strings.HasPrefix(location, fmt.Sprintf("file://%s", f.Base)) {
		return nil, fmt.Errorf("invalid file storage location: %s", location)
	}
	location = strings.TrimPrefix(location, "file://")

	return os.Open(location)
}

func (f FileStorage) Upload(ctx context.Context, checksum string, source string) (string, error) {
	file, err := os.Open(source)
	if err != nil {
		return "", err
	}
	// TODO: handle this error
	defer file.Close()
	return f.Create(ctx, checksum, file)
}
