create table if not exists actors
(
    id         uuid                                   not null
        constraint actors_pk
            primary key,
    actor_id   varchar                                not null
        constraint actors_actor_id
            unique,
    payload    jsonb                                  not null,
    created_at timestamp with time zone               not null,
    updated_at timestamp with time zone default now() not null
);

create table if not exists actor_aliases
(
    id         uuid                     not null
        constraint actor_subjects_pk
            primary key,
    actor_id   uuid                     not null,
    alias      varchar                  not null,
    created_at timestamp with time zone not null,
    updated_at timestamp with time zone not null,
    alias_type integer default 0        not null
);

create unique index if not exists actor_aliases_alias_uindex
    on actor_aliases (alias);

create table if not exists network_graph
(
    id                  uuid                     not null
        constraint network_graph_pk
            primary key,
    user_id             uuid                     not null,
    actor_id            uuid                     not null,
    activity            jsonb                    not null,
    relationship_type   integer default 0        not null,
    relationship_status integer default 0        not null,
    created_at          timestamp with time zone not null,
    updated_at          timestamp with time zone not null,
    constraint network_graph_user_actor_rel
        unique (user_id, actor_id, relationship_type)
);



create table if not exists actor_keys
(
    id         uuid                     not null
        constraint actor_keys_pk
            primary key,
    actor_id   uuid                     not null,
    key_id     varchar                  not null,
    pem        text                     not null,
    created_at timestamp with time zone not null,
    updated_at timestamp with time zone not null,
    constraint actor_keys_lookup
        unique (actor_id, key_id)
);



create table if not exists image_aliases
(
    id         uuid                     not null
        constraint image_aliases_pk primary key,
    image_id   uuid                     not null,
    -- TODO: make alias a sha256 value of the alias
    alias      varchar(200)             not null,
    created_at timestamp with time zone not null
);

create unique index if not exists image_aliases_pairs on image_aliases (image_id, alias);

create table if not exists images
(
    id           uuid              not null
        constraint asset_image_pk primary key,
    location     varchar(200)      not null,
    checksum     varchar(200)      not null,
    content_type integer default 0 not null,
    size         integer default 0 not null,
    height       integer default 0 not null,
    width        integer default 0 not null,
    blur         varchar(100)      not null,
    created_at   timestamp with time zone
);

create unique index if not exists asset_image_checksum_uindex on images (checksum);

create unique index if not exists asset_image_location_uindex on images (location);

create table if not exists peers
(
    id         uuid                     not null
        constraint peers_pk primary key,
    inbox      varchar(100)             not null,
    created_at timestamp with time zone not null
);

create unique index if not exists peers_inbox_uindex on peers (inbox);

create table if not exists users
(
    id               uuid                                         not null
        constraint users_pkey primary key,
    email            varchar(100)
        constraint users_email_key unique,
    password         bytea,
    created_at       timestamp with time zone                     not null,
    updated_at       timestamp with time zone                     not null,
    last_auth_at     timestamp with time zone,
    location         varchar(48) default 'UTC'::character varying not null, -- user time zone
    locale           varchar(8)  default 'en'::character varying  not null, -- user language
    public_key       text                                         not null,
    private_key      text                                         not null,
    name             varchar(36)                                  not null,
    display_name     varchar(100)                                 not null,
    about            text                                         not null,
    accept_followers boolean     default true                     not null
);

create unique index if not exists users_username_uindex on users (name);

create table objects
(
    id         uuid
        constraint objects_pk
            primary key,
    created_at timestamp with time zone not null,
    updated_at timestamp with time zone not null,
    payload    jsonb                    not null,
    object_id  varchar                  not null
);

create unique index objects_object_uindex on objects (object_id);

ALTER TABLE objects
    ADD CONSTRAINT objects_object_uindex UNIQUE USING INDEX objects_object_uindex;

create table object_events
(
    id          uuid
        constraint object_events_pk primary key,
    created_at  timestamp with time zone not null,
    updated_at  timestamp with time zone not null,
    activity_id varchar                  not null,
    object_id   uuid                     not null,
    payload     jsonb                    not null
);

create unique index object_events_activity_uindex on object_events (activity_id);

ALTER TABLE object_events
    ADD CONSTRAINT object_events_activity_uindex UNIQUE USING INDEX object_events_activity_uindex;

create table user_object_events
(
    id          uuid
        constraint user_object_events_pk primary key,
    created_at  timestamp with time zone not null,
    updated_at  timestamp with time zone not null,
    user_id     uuid                     not null,
    activity_id uuid                     not null,
    object_id   uuid                     not null,
    public      bool                     not null default false
);

create unique index user_object_events_user_activity_uindex on user_object_events (user_id, activity_id, object_id);

ALTER TABLE user_object_events
    ADD CONSTRAINT user_object_events_user_activity_uindex UNIQUE USING INDEX user_object_events_user_activity_uindex;

create table user_object_tags
(
    id          uuid
        constraint user_object_tags_pk primary key,
    created_at  timestamp with time zone not null,
    updated_at  timestamp with time zone not null,
    user_id     uuid                     not null,
    activity_id uuid                     not null,
    object_id   uuid                     not null,
    tag         varchar                  not null
);

create unique index user_object_tags_tagged_uindex on user_object_tags (user_id, object_id, tag);

ALTER TABLE user_object_tags
    ADD CONSTRAINT user_object_tags_tagged_uindex UNIQUE USING INDEX user_object_tags_tagged_uindex;

create table user_feed
(
    id          uuid                     not null
        constraint user_feed_pk
            primary key,
    activity_id uuid                     not null,
    object_id   uuid                     not null,
    user_id     uuid                     not null,
    created_at  timestamp with time zone not null
);

create unique index user_feed_activity_uindex on user_feed (user_id, activity_id, object_id);

ALTER TABLE user_feed
    ADD CONSTRAINT user_feed_activity_uindex UNIQUE USING INDEX user_feed_activity_uindex;