package storage

import (
	"context"
	"database/sql"
	"time"

	"github.com/gofrs/uuid"
	_ "github.com/lib/pq"
	"go.uber.org/zap"

	"github.com/ngerakines/tavern/errors"
)

type Storage interface {
	RowCount(ctx context.Context, query string, args ...interface{}) (int, error)

	UserStorage
	ActorStorage
	FollowerStorage
	PeerStorage
	ObjectStorage
	AssetStorage
	InstanceStatsStorage

	GetExecutor() QueryExecute
	GetLogger() *zap.Logger
}

type QueryExecute interface {
	ExecContext(ctx context.Context, query string, args ...interface{}) (sql.Result, error)
	QueryContext(ctx context.Context, query string, args ...interface{}) (*sql.Rows, error)
	QueryRowContext(ctx context.Context, query string, args ...interface{}) *sql.Row
}

type Count struct {
	Key   string
	Count int
}

func DefaultStorage(db *sql.DB, logger *zap.Logger) Storage {
	return pgStorage{
		db:     db,
		now:    defaultNowFunc,
		logger: logger,
	}
}

type pgStorage struct {
	db     QueryExecute
	now    nowFunc
	logger *zap.Logger
}

type transactionScopedWork func(db QueryExecute) error
type transactionScopedStorage func(storage Storage) error
type nowFunc func() time.Time

func defaultNowFunc() time.Time {
	return time.Now().UTC()
}

func runTransactionWithOptions(db QueryExecute, txBody transactionScopedWork) error {
	realDB, ok := db.(*sql.DB)
	if !ok {
		return txBody(db)
	}

	tx, err := realDB.Begin()
	if err != nil {
		return errors.NewDatabaseTransactionFailedError(err)
	}

	err = txBody(tx)
	if err != nil {
		if txErr := tx.Rollback(); txErr != nil {
			return errors.NewDatabaseTransactionFailedError(txErr)
		}
		return err
	}
	return errors.WrapDatabaseTransactionFailedError(tx.Commit())
}

func TransactionalStorage(ctx context.Context, storage Storage, txBody transactionScopedStorage) error {
	executor := storage.GetExecutor()
	realDB, ok := executor.(*sql.DB)
	if !ok {
		return txBody(storage)
	}

	tx, err := realDB.BeginTx(ctx, nil)
	if err != nil {
		return errors.NewDatabaseTransactionFailedError(err)
	}

	err = txBody(pgStorage{
		db:     tx,
		now:    defaultNowFunc,
		logger: storage.GetLogger(),
	})
	if err != nil {
		if txErr := tx.Rollback(); txErr != nil {
			return errors.NewDatabaseTransactionFailedError(txErr)
		}
		return err
	}
	return errors.WrapDatabaseTransactionFailedError(tx.Commit())
}

func (s pgStorage) GetExecutor() QueryExecute {
	return s.db
}

func (s pgStorage) GetLogger() *zap.Logger {
	return s.logger
}

func (s pgStorage) RowCount(ctx context.Context, query string, args ...interface{}) (int, error) {
	return s.wrappedRowCount(errors.WrapQueryFailedError, s.db, ctx, query, args...)
}

func (s pgStorage) rowCount(qec QueryExecute, ctx context.Context, query string, args ...interface{}) (int, error) {
	return s.wrappedRowCount(errors.WrapQueryFailedError, qec, ctx, query, args...)
}

func (s pgStorage) wrappedRowCount(ew errors.ErrorWrapper, qec QueryExecute, ctx context.Context, query string, args ...interface{}) (int, error) {
	s.logger.Debug("row count query", zap.String("query", query), zap.Reflect("args", args))

	var total int

	err := qec.QueryRowContext(ctx, query, args...).Scan(&total)
	if err != nil {
		return -1, ew(err)
	}
	return total, nil
}

func (s pgStorage) wrappedExists(ew errors.ErrorWrapper, ctx context.Context, query string, args ...interface{}) (bool, error) {
	s.logger.Debug("row count query", zap.String("query", query), zap.Reflect("args", args))

	var total int
	err := s.db.QueryRowContext(ctx, query, args...).Scan(&total)
	if err != nil {
		return false, ew(err)
	}
	return total != 0, nil
}

func (s pgStorage) keyedCount(ew errors.ErrorWrapper, ctx context.Context, query string, args ...interface{}) ([]Count, error) {
	results := make([]Count, 0)
	rows, err := s.db.QueryContext(ctx, query, args...)
	if err != nil {
		return nil, ew(err)
	}
	defer rows.Close()
	for rows.Next() {
		var count Count
		if err := rows.Scan(&count.Key, &count.Count); err != nil {
			return nil, ew(err)
		}
		results = append(results, count)
	}
	return results, nil
}

func (s pgStorage) keysToUUID(ew errors.ErrorWrapper, ctx context.Context, query string, args ...interface{}) (map[string]uuid.UUID, error) {
	results := make(map[string]uuid.UUID)
	rows, err := s.db.QueryContext(ctx, query, args...)
	if err != nil {
		return nil, ew(err)
	}
	defer rows.Close()
	for rows.Next() {
		var key string
		var id uuid.UUID
		if err := rows.Scan(&key, &id); err != nil {
			return nil, ew(err)
		}
		results[key] = id
	}
	return results, nil
}

func (s pgStorage) uuidsToPayload(ew errors.ErrorWrapper, ctx context.Context, query string, args ...interface{}) (map[uuid.UUID]Payload, error) {
	results := make(map[uuid.UUID]Payload)
	rows, err := s.db.QueryContext(ctx, query, args...)
	if err != nil {
		return nil, ew(err)
	}
	defer rows.Close()
	for rows.Next() {
		var key uuid.UUID
		var payload Payload
		if err := rows.Scan(&key, &payload); err != nil {
			return nil, ew(err)
		}
		results[key] = payload
	}
	return results, nil
}

func (s pgStorage) uuidsToUuids(ew errors.ErrorWrapper, ctx context.Context, query string, args ...interface{}) (map[uuid.UUID]uuid.UUID, error) {
	results := make(map[uuid.UUID]uuid.UUID)
	rows, err := s.db.QueryContext(ctx, query, args...)
	if err != nil {
		return nil, ew(err)
	}
	defer rows.Close()
	for rows.Next() {
		var key uuid.UUID
		var value uuid.UUID
		if err := rows.Scan(&key, &value); err != nil {
			return nil, ew(err)
		}
		results[key] = value
	}
	return results, nil
}

func (s pgStorage) toUUIDMultiMap(ew errors.ErrorWrapper, ctx context.Context, query string, args ...interface{}) (map[uuid.UUID][]uuid.UUID, error) {
	results := make(map[uuid.UUID][]uuid.UUID)
	rows, err := s.db.QueryContext(ctx, query, args...)
	if err != nil {
		return nil, ew(err)
	}
	defer rows.Close()
	for rows.Next() {
		var key uuid.UUID
		var value uuid.UUID
		if err := rows.Scan(&key, &value); err != nil {
			return nil, ew(err)
		}
		values, hasValues := results[key]
		if !hasValues {
			values = make([]uuid.UUID, 0)
		}
		results[key] = append(values, value)
	}
	return results, nil
}
