package web

import (
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"strings"
	"time"

	"github.com/gin-gonic/gin"
	"github.com/gofrs/uuid"
	"github.com/kr/pretty"
	"github.com/yukimochi/httpsig"
	"go.uber.org/zap"

	"github.com/ngerakines/tavern/errors"
	"github.com/ngerakines/tavern/fed"
	"github.com/ngerakines/tavern/storage"
)

const maxInboxRequestSize = 1 * 1024 * 1024

func (h handler) actorInbox(c *gin.Context) {
	name := c.Param("name")

	user, err := h.storage.GetUserByName(c.Request.Context(), name)
	if err != nil {
		if errors.Is(err, errors.NewNotFoundError(nil)) {
			h.notFoundJSON(c, nil)
			return
		}
		h.internalServerErrorJSON(c, err, zap.String("name", name))
		return
	}

	defer c.Request.Body.Close()

	body, err := ioutil.ReadAll(io.LimitReader(c.Request.Body, maxInboxRequestSize+1))
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}
	if len(body) > maxInboxRequestSize {
		h.writeJSONError(c, http.StatusRequestEntityTooLarge, err)
		return
	}

	fmt.Println(string(body))

	payload, err := storage.PayloadFromBytes(body)
	if err != nil {
		h.badRequestJSON(c, err)
		return
	}

	pretty.Println(payload)

	if skipActorInbox(payload) {
		h.logger.Debug("actor inbox can ignore message", zap.String("user", name))
		c.Status(http.StatusOK)
		return
	}

	payloadType, _ := storage.JSONString(payload, "type")

	actor, _ := storage.JSONString(payload, "actor")
	if len(actor) > 0 {
		err = h.webFingerQueue.Add(actor)
		if err != nil {
			h.logger.Error("unable to add actor to web finger queue", zap.String("actor", actor))
		}
	}

	switch payloadType {
	case "Follow":
		h.actorInboxFollow(c, user, payload, body)
	case "Undo":
		h.actorInboxUndo(c, user, payload)
	case "Accept":
		h.actorInboxAccept(c, user, payload)
	case "Reject":
		h.actorInboxReject(c, user, payload)
	case "Create":
		h.actorInboxCreate(c, user, payload)
	case "Announce":
		h.actorInboxAnnounce(c, user, payload)
	case "Delete":
		h.actorInboxDelete(c, user, payload)
	default:
		h.logger.Warn("User received unexpected payload type", zap.String("type", payloadType), zap.String("user", name))
		c.Status(http.StatusOK)
	}
}

func (h handler) actorInboxAccept(c *gin.Context, user *storage.User, payload storage.Payload) {
	if err := h.verifySignature(c); err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	userActorID := storage.NewActorID(user.Name, h.domain)

	self, hasActor := storage.JSONDeepString(payload, "object", "actor")
	target, hasTarget := storage.JSONDeepString(payload, "object", "object")

	if !hasActor || !hasTarget || self != string(userActorID) {
		h.logger.Warn("unable to process accept request",
			zap.String("self", self),
			zap.String("actor", target),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	remoteActorID, err := h.storage.ActorRowIDForActorID(c.Request.Context(), target)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	err = h.storage.UpdateFollowingAccepted(c.Request.Context(), user.ID, remoteActorID)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	c.Status(http.StatusOK)
}

func (h handler) actorInboxReject(c *gin.Context, user *storage.User, payload storage.Payload) {
	if err := h.verifySignature(c); err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	userActorID := storage.NewActorID(user.Name, h.domain)

	obj, hasObject := storage.JSONMap(payload, "object")
	self, hasActor := storage.JSONString(obj, "actor")
	target, hasTarget := storage.JSONString(obj, "object")

	if !hasObject || !hasActor || !hasTarget || self != string(userActorID) {
		h.logger.Warn("unable to process reject request",
			zap.String("self", self),
			zap.String("actor", target),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	remoteActorID, err := h.storage.ActorRowIDForActorID(c.Request.Context(), target)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	err = h.storage.UpdateFollowingRejected(c.Request.Context(), user.ID, remoteActorID)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	c.Status(http.StatusOK)
}

func (h handler) actorInboxFollow(c *gin.Context, user *storage.User, payload storage.Payload, body []byte) {
	ctx := c.Request.Context()

	userActor, err := h.storage.GetActor(ctx, user.ActorID)
	if err != nil {
		h.hardFail(c, err)
		return
	}

	userActorID := storage.NewActorID(user.Name, h.domain)

	self, hasActor := storage.JSONString(payload, "object")
	target, hasTarget := storage.JSONString(payload, "actor")

	if !hasActor || !hasTarget || self != string(userActorID) {
		h.logger.Warn("unable to process follow request",
			zap.String("object", self),
			zap.String("actor", target),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	followerActor, err := fed.GetOrFetchActor(ctx, h.storage, h.logger, h.httpClient, target)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	if err := h.verifySignature(c); err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	remoteActorID, err := h.storage.ActorRowIDForActorID(c.Request.Context(), target)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	err = h.storage.CreatePendingFollower(ctx, user.ID, remoteActorID, payload)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	if user.AcceptFollowers {

		response := storage.EmptyPayload()
		acceptActivityID := storage.NewV4()
		response["@context"] = "https://www.w3.org/ns/activitystreams"
		response["actor"] = storage.NewActorID(user.Name, h.domain)
		response["id"] = fmt.Sprintf("https://%s/activity/%s", h.domain, acceptActivityID)
		response["object"] = payload
		response["to"] = followerActor.GetID()
		response["type"] = "Accept"
		response["published"] = time.Now().UTC().Format("2006-01-02T15:04:05Z")
		responsePayload := response.Bytes()

		nc := fed.ActorClient{
			HTTPClient: h.httpClient,
			Logger:     h.logger,
		}
		err = nc.SendToInbox(ctx, h.userActor(user, userActor), followerActor, responsePayload)
		if err != nil {
			h.internalServerErrorJSON(c, err)
			return
		}

		err = h.storage.UpdateFollowerApproved(ctx, user.ID, remoteActorID)
		if err != nil {
			h.internalServerErrorJSON(c, err)
			return
		}
	}

	c.Status(http.StatusOK)
}

func (h handler) actorInboxUndoFollow(c *gin.Context, user *storage.User, payload storage.Payload) {
	ctx := c.Request.Context()

	userActorID := storage.NewActorID(user.Name, h.domain)

	self, hasActor := storage.JSONDeepString(payload, "object", "object")
	target, hasTarget := storage.JSONDeepString(payload, "object", "actor")

	if !hasActor || !hasTarget || self != string(userActorID) {
		h.logger.Warn("unable to process follow request",
			zap.String("object", self),
			zap.String("actor", target),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	if err := h.verifySignature(c); err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	remoteActorID, err := h.storage.ActorRowIDForActorID(c.Request.Context(), target)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	err = h.storage.RemoveFollower(ctx, user.ID, remoteActorID)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	c.Status(http.StatusOK)
}

func (h handler) actorInboxUndo(c *gin.Context, user *storage.User, payload storage.Payload) {
	innerType, _ := storage.JSONDeepString(payload, "object", "type")
	switch innerType {
	case "Follow":
		h.actorInboxUndoFollow(c, user, payload)
	default:
		h.logger.Warn("User received unexpected undo payload type", zap.String("type", innerType), zap.String("user", user.Name))
		c.Status(http.StatusOK)
	}
}

func (h handler) actorInboxCreate(c *gin.Context, user *storage.User, payload storage.Payload) {
	isRelevant, err := h.isActivityRelevant(c.Request.Context(), payload, user)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}
	if !isRelevant {
		c.Status(http.StatusOK)
		return
	}

	if err = h.verifySignature(c); err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	activityID, hasActivityID := storage.JSONString(payload, "id")
	actorID, hasActorID := storage.JSONString(payload, "actor")
	activityObjectID, hasActivityObjectID := storage.JSONDeepString(payload, "object", "id")
	activityObject, hasActivityObject := storage.JSONMap(payload, "object")
	activityObjectType, hasActivityObjectType := storage.JSONDeepString(payload, "object", "type")

	if !hasActivityID || !hasActorID || !hasActivityObjectID || !hasActivityObject || !hasActivityObjectType {
		h.logger.Info("ignoring invalid create note activity",
			zap.String("id", activityID),
			zap.String("actor", actorID),
			zap.String("object_id", activityObjectID),
			zap.String("object_type", activityObjectType),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}
	if activityObjectType != "Note" {
		h.logger.Warn("unknown create object type",
			zap.String("id", activityID),
			zap.String("actor", actorID),
			zap.String("object_id", activityObjectID),
			zap.String("object_type", activityObjectType),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	inReplyTo, hasInReplyTo := storage.JSONDeepString(payload, "object", "inReplyTo")

	// localObjectPrefix := fmt.Sprintf("https://%s/object/", h.domain)

	ctx := c.Request.Context()

	err = storage.TransactionalStorage(ctx, h.storage, func(storage storage.Storage) error {
		activityObjectRowID, err := storage.RecordObject(ctx, activityObject, activityObjectID)
		if err != nil {
			return err
		}
		activityRowID, err := storage.RecordObjectEvent(ctx, activityID, activityObjectRowID, payload)
		if err != nil {
			return err
		}
		_, err = storage.RecordUserFeed(ctx, activityRowID, activityObjectRowID, user.ID)
		if err != nil {
			return err
		}
		if hasInReplyTo  {
			replyRowID, err := storage.ObjectRowIDForObjectID(ctx, inReplyTo)
			if err != nil {
				return err
			}
			_, err = storage.RecordObjectReply(ctx, activityObjectRowID, replyRowID)
			if err != nil {
				return err
			}
		}
		return nil
	})
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	if err = h.crawlQueue.Add(strings.Join([]string{user.ID.String(), activityObjectID}, ",")); err != nil {
		h.logger.Warn("error queueing crawl work", zap.Error(err))
	}

	obj, ok := storage.JSONMap(activityObject, "object")
	if ok {
		attachments, ok := storage.JSONMapList(obj, "attachment")
		if ok {
			for _, attachment := range attachments {
				mediaType, hasMediaType := storage.JSONString(attachment, "mediaType")
				url, hasURL := storage.JSONString(attachment, "url")
				if hasMediaType && hasURL && strings.HasPrefix(url, "https://") {
					if mediaType == "image/jpeg" || mediaType == "image/png" {
						if err = h.assetQueue.Add(url); err != nil {
							h.logger.Warn("error queueing asset download work", zap.Error(err))
						}
					}
				}
			}
		}
	}

	c.Status(http.StatusOK)
}

func (h handler) actorInboxAnnounce(c *gin.Context, user *storage.User, payload storage.Payload) {
	isRelevant, err := h.isActivityRelevant(c.Request.Context(), payload, user)
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}
	if !isRelevant {
		c.Status(http.StatusOK)
		return
	}

	if err = h.verifySignature(c); err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	userActor, err := h.storage.GetActor(c.Request.Context(), user.ActorID)
	if err != nil {
		h.hardFail(c, err)
		return
	}

	activityID, hasActivityID := storage.JSONString(payload, "id")
	actorID, hasActorID := storage.JSONString(payload, "actor")
	activityObjectID, hasActivityObjectID := storage.JSONString(payload, "object")

	if !hasActivityID || !hasActorID || !hasActivityObjectID {
		h.logger.Info("ignoring invalid announce activity",
			zap.String("id", activityID),
			zap.String("actor", actorID),
			zap.String("object_id", activityObjectID),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	ctx := c.Request.Context()

	localActivityPrefix := fmt.Sprintf("https://%s/activity/", h.domain)
	if strings.HasPrefix(activityObjectID, localActivityPrefix) {

		activityObjectRowID, err := uuid.FromString(strings.TrimPrefix(activityObjectID, localActivityPrefix))
		if err != nil {
			h.internalServerErrorJSON(c, err)
			return
		}

		err = storage.TransactionalStorage(ctx, h.storage, func(storage storage.Storage) error {
			activityRowID, err := storage.RecordObjectEvent(ctx, activityID, activityObjectRowID, payload)
			if err != nil {
				return err
			}
			_, err = storage.RecordUserFeed(ctx, activityRowID, activityObjectRowID, user.ID)
			if err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			h.internalServerErrorJSON(c, err)
			return
		}

		c.Status(http.StatusOK)
		return
	}

	ac := fed.ActivityClient{
		HTTPClient: h.httpClient,
		Logger:     h.logger,
	}

	_, objectPayload, err := ac.GetSigned(activityObjectID, h.userActor(user, userActor))
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	objectID, hasObjectID := storage.JSONString(objectPayload, "id")
	if !hasObjectID || objectID != activityObjectID {
		h.logger.Info("ignoring invalid announce activity",
			zap.String("id", activityID),
			zap.String("actor", actorID),
			zap.Strings("object_id", []string{activityObjectID, objectID}),
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}
	err = storage.TransactionalStorage(ctx, h.storage, func(storage storage.Storage) error {
		activityObjectRowID, err := storage.RecordObject(ctx, objectPayload, activityObjectID)
		if err != nil {
			return err
		}
		activityRowID, err := storage.RecordObjectEvent(ctx, activityID, activityObjectRowID, payload)
		if err != nil {
			return err
		}
		_, err = storage.RecordUserFeed(ctx, activityRowID, activityObjectRowID, user.ID)
		if err != nil {
			return err
		}
		return nil
	})
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	if err = h.crawlQueue.Add(strings.Join([]string{user.ID.String(), objectID}, ",")); err != nil {
		h.logger.Warn("error queueing crawl work", zap.Error(err))
	}

	c.Status(http.StatusOK)
}

func (h handler) actorInboxDelete(c *gin.Context, user *storage.User, payload storage.Payload) {
	err := h.verifySignature(c)
	if err != nil {
		h.unauthorizedJSON(c, err)
		return
	}

	ctx := c.Request.Context()

	var objectID string
	var objectRowID uuid.UUID
	foundObjectID := false
	objectIDs := storage.CollectJSONDeepStrings(payload, []string{"object"}, []string{"object", "id"})
	for _, objectID = range objectIDs {
		objectRowID, err = h.storage.ObjectRowIDForObjectID(ctx, objectID)
		if err == nil {
			foundObjectID = true
			break
		}
	}

	if !foundObjectID {
		h.logger.Warn("unable to process delete request",
			zap.String("user", user.Name),
		)
		c.Status(http.StatusOK)
		return
	}

	activityURL, _ := storage.JSONString(payload, "id")
	var tombstone storage.Payload
	if object, hasObject := storage.JSONMap(payload, "object"); hasObject {
		tombstone = object
	}
	if tombstone == nil {
		// TODO: Replace this with a fetch of the object to get a proper tombstone from the remote instance.
		tombstone = storage.EmptyPayload()
		tombstone["type"] = "Tombstone"
		// TODO: Ensure that the object is, in fact, a note.
		tombstone["formerType"] = "Note"
		tombstone["id"] = objectID
		now := time.Now().UTC()
		deletedAt := now.Format("2006-01-02T15:04:05Z")
		tombstone["deleted"] = deletedAt
	}

	txErr := storage.TransactionalStorage(ctx, h.storage, func(storage storage.Storage) error {
		err = storage.UpdateObjectPayload(ctx, objectRowID, tombstone)
		if err != nil {
			return err
		}

		_, err = storage.RecordObjectEvent(ctx, activityURL, objectRowID, payload)
		if err != nil {
			return err
		}

		return nil
	})
	if txErr != nil {
		h.internalServerErrorJSON(c, txErr)
		return
	}

	c.Status(http.StatusOK)
}

func skipActorInbox(j storage.Payload) bool {
	t, _ := storage.JSONString(j, "type")
	a, _ := storage.JSONString(j, "actor")
	o, _ := storage.JSONString(j, "object")
	if t == "Delete" && a == o {
		return true
	}
	return false
}

func (h handler) isActivityRelevant(ctx context.Context, activity storage.Payload, user *storage.User) (bool, error) {
	actor, ok := storage.JSONString(activity, "actor")
	if ok {

		if err := h.webFingerQueue.Add(actor); err != nil {
			h.logger.Warn("unable to add actor to webfinger queue",
				zap.Error(err),
				zap.String("actor", actor))
		}

		remoteActorID, err := h.storage.ActorRowIDForActorID(ctx, actor)
		if err != nil {
			if errors.Is(err, errors.NewNotFoundError(nil)) {
				return false, nil
			}
			h.logger.Debug("error looking up actor to see if activity is relevant",
				zap.String("user", user.Name),
				zap.String("actor", actor),
				zap.Error(err))
			return false, err
		}

		isFollowing, err := h.storage.IsFollowing(ctx, user.ID, remoteActorID)
		if err != nil {
			h.logger.Debug("error determining if user follows actor",
				zap.String("user", user.Name),
				zap.String("actor", actor),
				zap.Error(err))
			return false, err
		}
		if isFollowing {
			return true, nil
		}
	}

	actorID := storage.NewActorID(user.Name, h.domain)

	obj, ok := storage.JSONMap(activity, "object")
	if ok {
		to, ok := storage.JSONStrings(obj, "to")
		if ok {
			for _, i := range to {
				if i == string(actorID) {
					return true, nil
				}
			}
		}
		cc, ok := storage.JSONStrings(obj, "cc")
		if ok {
			for _, i := range cc {
				if i == string(actorID) {
					return true, nil
				}
			}
		}
		tag, ok := storage.JSONMapList(obj, "tag")
		if ok {
			for _, i := range tag {
				tagType, hasTagType := storage.JSONString(i, "type")
				href, hasHref := storage.JSONString(i, "href")
				if hasTagType && hasHref && tagType == "Mention" && href == string(actorID) {
					return true, nil
				}
			}
		}
	}

	return false, nil
}

func (h handler) verifySignature(c *gin.Context) error {
	// Host header isn't set for some reason.
	c.Request.Header.Add("Host", c.Request.Host)
	verifier, err := httpsig.NewVerifier(c.Request)
	if err != nil {
		return err
	}
	key, err := h.storage.GetKey(c.Request.Context(), verifier.KeyId())
	if err != nil {
		return err
	}
	publicKey, err := key.GetDecodedPublicKey()
	if err != nil {
		return err
	}
	if err = verifier.Verify(publicKey, httpsig.RSA_SHA256); err != nil {
		return err
	}
	return nil
}
