package web

import (
	"net/http"

	"github.com/gin-contrib/sessions"
	"github.com/gin-gonic/gin"
	"go.uber.org/zap"

	"github.com/ngerakines/tavern/errors"
	"github.com/ngerakines/tavern/storage"
)

func (h handler) configure(c *gin.Context) {
	session := sessions.Default(c)
	ctx := c.Request.Context()
	trans, transOK := c.Get("trans")
	if !transOK {
		panic("trans not found in context")
	}
	data := gin.H{
		"flashes": getFlashes(session),
		"Trans":   trans,
	}

	user, err := h.storage.GetUserBySession(ctx, session)
	if err != nil {
		if errors.Is(err, errors.UserSessionNotFoundError{}) {

			if err = appendFlashError(session, "Not Authenticated"); err != nil {
				h.hardFail(c, errors.NewCannotSaveSessionError(err))
				return
			}

			c.Redirect(http.StatusFound, "/")
			return
		}
		h.hardFail(c, err)
		return
	}

	data["user"] = user
	data["authenticated"] = true

	data["fed_config"] = map[string]bool{
		"reply_collection_updates": h.fedConfig.AllowReplyCollectionUpdates,
		"auto_accept_followers":    h.fedConfig.AllowAutoAcceptFollowers,
	}

	data["user_config"] = map[string]bool{
		"reply_collection_updates": user.ReplyCollectionUpdates,
		"auto_accept_followers":    user.AcceptFollowers,
	}

	if err = session.Save(); err != nil {
		h.hardFail(c, err)
		return
	}

	c.HTML(http.StatusOK, "configure", data)
}

func (h handler) saveUserSettings(c *gin.Context) {
	session := sessions.Default(c)
	ctx := c.Request.Context()
	user, err := h.storage.GetUserBySession(ctx, session)
	if err != nil {
		h.hardFail(c, err)
		return
	}

	h.logger.Info("saving user settings",
		zap.String("auto_accept_followers", c.PostForm("auto_accept_followers")),
		zap.String("reply_collection_updates", c.PostForm("reply_collection_updates")),
	)

	err = storage.TransactionalStorage(ctx, h.storage, func(storage storage.Storage) error {
		txErr := storage.UpdateUserAutoAcceptFollowers(ctx, user.ID, c.PostForm("auto_accept_followers") == "checked")
		if txErr != nil {
			return txErr
		}
		return storage.UpdateUserReplyCollectionUpdates(ctx, user.ID, c.PostForm("reply_collection_updates") == "checked")
	})
	if err != nil {
		h.hardFail(c, err)
		return
	}

	c.Redirect(http.StatusFound, h.url("configure"))
}
