package web

import (
	"encoding/hex"
	"fmt"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/gin-contrib/sessions"
	"github.com/gin-gonic/gin"
	"github.com/gofrs/uuid"
	"go.uber.org/zap"

	"github.com/ngerakines/tavern/common"
	"github.com/ngerakines/tavern/errors"
	"github.com/ngerakines/tavern/storage"
)

func (h handler) viewUser(c *gin.Context) {
	session := sessions.Default(c)
	ctx := c.Request.Context()
	trans, transOK := c.Get("trans")
	if !transOK {
		panic("trans not found in context")
	}
	data := gin.H{
		"flashes": getFlashes(session),
		"Trans":   trans,
	}

	authenticated := true
	localUser, err := h.storage.GetUserBySession(ctx, session)
	if err != nil {
		if !errors.Is(err, errors.NewNotFoundError(nil)) {
			h.internalServerErrorJSON(c, err)
			return
		}
		authenticated = false
	}

	data["authenticated"] = authenticated
	data["user"] = localUser

	name := c.Param("name")

	viewUser, err := h.storage.GetUserByName(ctx, name)
	if err != nil {
		if errors.Is(err, errors.NewNotFoundError(nil)) {
			h.notFoundJSON(c, err)
			return
		}
		h.internalServerErrorJSON(c, err)
		return
	}

	data["view_user"] = viewUser
	data["feed_template"] = "profile"
	data["self_url"] = h.url("profile", viewUser.Name)

	page := intParam(c, "page", 1)
	limit := 20

	h.displayObjectFeed(c, true, data, func(user *storage.User) (i int, payloads []storage.Payload, err error) {
		total, err := h.storage.CountObjectPayloadsInUserOutbox(ctx, viewUser.ID)
		if err != nil {
			return 0, nil, err
		}
		h.logger.Debug("counted objects in feed", zap.Int("total", total))
		objects, err := h.storage.ListObjectPayloadsInUserOutbox(ctx, viewUser.ID, limit, (page-1)*limit)
		if err != nil {
			return 0, nil, err
		}
		h.logger.Debug("queried objects", zap.Int("count", len(objects)))

		return total, objects, nil
	})
}

func (h handler) viewObject(c *gin.Context) {
	data, localUser, _ /* session */, cont := h.loggedIn(c, true)
	if !cont {
		return
	}
	ctx := c.Request.Context()

	localUserActor, err := h.storage.GetActor(c.Request.Context(), localUser.ActorID)
	if err != nil {
		h.hardFail(c, err)
		return
	}

	objectID := common.ObjectURL(h.domain, c.Param("object"))

	objectRowID, err := h.storage.ObjectRowIDForObjectID(ctx, objectID)
	if err != nil {
		h.hardFail(c, err)
		return
	}

	object, err := h.storage.ObjectPayloadByObjectID(ctx, objectID)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	data["object"] = object

	objectParents, err := h.storage.ObjectParentsByObjectID(ctx, objectRowID)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	data["parents"] = objectParents

	objectChildren, err := h.storage.ObjectChildrenByObjectID(ctx, objectRowID)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	data["children"] = objectChildren

	var topParent uuid.UUID

	actorNames := make(map[string]string)
	actorIcons := make(map[string]string)
	objectIDToObjectRowID := make(map[string]uuid.UUID)

	allChildren := make(map[uuid.UUID][]uuid.UUID)

	objectRowIDs := make([]uuid.UUID, 0)
	for k, v := range objectParents {
		objectRowIDs = append(objectRowIDs, k)
		objectRowIDs = append(objectRowIDs, v...)
		for _, vc := range v {

			allChildrenValues, hasAllChildrenValues := allChildren[vc]
			if !hasAllChildrenValues {
				allChildrenValues = make([]uuid.UUID, 0)
			}
			allChildren[vc] = append(allChildrenValues, k)

			if _, ok := objectParents[vc]; !ok {
				topParent = vc
			}
		}
	}
	for k, v := range objectChildren {
		objectRowIDs = append(objectRowIDs, k)
		objectRowIDs = append(objectRowIDs, v...)

		allChildrenValues, hasAllChildrenValues := allChildren[k]
		if !hasAllChildrenValues {
			allChildrenValues = make([]uuid.UUID, 0)
		}
		allChildren[k] = append(allChildrenValues, v...)
	}

	data["top_parent"] = topParent
	data["all_children"] = allChildren

	refs, err := h.storage.ObjectPayloadByUUID(ctx, objectRowIDs)
	if err != nil {
		h.hardFail(c, err)
		return
	}

	collectedActorIDs := make([]string, 0)

	for orid, p := range refs {
		objectID, hasObjectID := storage.JSONString(p, "id")
		if hasObjectID {
			objectIDToObjectRowID[objectID] = orid
		}
		attributedTo, hasAttributedTo := storage.JSONString(p, "attributedTo")
		if hasAttributedTo {
			collectedActorIDs = append(collectedActorIDs, attributedTo)
		}
	}

	data["refs"] = refs

	actors, err := h.storage.ActorsByActorID(ctx, collectedActorIDs)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	var actorRowIDs []uuid.UUID
	for _, actor := range actors {
		actorRowIDs = append(actorRowIDs, actor.ID)
	}
	actorSubjects, err := h.storage.ActorSubjects(ctx, actorRowIDs)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	allActors := h.gatherActors(actors, actorSubjects)
	for actorID, actorSummary := range allActors {
		actorNames[actorID] = actorSummary["at"]
		actorIcons[actorID] = actorSummary["icon"]
	}

	boosts, err := h.storage.CountObjectBoostsByObjectIDs(ctx, objectRowIDs)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	mappedBoosts := make(map[string]int)
	for _, boost := range boosts {
		mappedBoosts[boost.Key] = boost.Count
	}

	userBoosts, err := h.storage.CountObjectBoostsByActorObjectIDs(ctx, localUserActor.ID, objectRowIDs)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	mappedUserBoosts := make(map[string]int)
	for _, userBoost := range userBoosts {
		mappedUserBoosts[userBoost.Key] = userBoost.Count
	}

	replies, err := h.storage.CountObjectRepliesByObjectIDs(ctx, objectRowIDs)
	if err != nil {
		h.hardFail(c, err)
		return
	}
	mappedReplies := make(map[string]int)
	for _, reply := range replies {
		mappedReplies[reply.Key] = reply.Count
	}

	if topParent == uuid.Nil {
		topParent = objectRowID
	}

	topParentRef, hasTopParentRef := refs[topParent]
	if !hasTopParentRef {
		c.HTML(http.StatusOK, "object", data)
		return
	}

	// addViewObject(root *viewObject, parentObjectRowID, objectRowID uuid.UUID, payload storage.Payload)
	root := addViewObject(nil, uuid.Nil, topParent, topParentRef)
	root.populate(0, allChildren, refs)

	mediaPrefix := fmt.Sprintf("https://%s/asset/image/", h.domain)

	root.applyView(func(viewObjectRowID uuid.UUID, objectPayload storage.Payload, viewContext *viewObjectContext) *viewObjectContext {
		objectID, hasObjectID := storage.JSONString(objectPayload, "id")
		if !hasObjectID {
			return &viewObjectContext{Error: "Object has no id."}
		}
		viewContext.ObjectID = objectID

		attributedTo, hasAttributedTo := storage.JSONString(objectPayload, "attributedTo")
		content, hasContent := storage.JSONString(objectPayload, "content")
		if !hasAttributedTo {
			return &viewObjectContext{Error: "Object has no attribution."}
		}
		if !hasContent {
			return &viewObjectContext{Error: "Object has no content."}
		}

		viewContext.Icon = fmt.Sprintf("https://%s/avatar/png/unknown/unknown?size=60", h.domain)

		if actorIcon, hasActorIcon := actorIcons[attributedTo]; hasActorIcon {
			viewContext.Icon = actorIcon
		}

		destinations := storage.ActivityDestinations(objectPayload)
		for _, destination := range destinations {
			if destination == "https://www.w3.org/ns/activitystreams#Public" {
				viewContext.Public = true
				break
			}
		}

		viewContext.Content = content

		viewContext.AttributedTo = attributedTo
		viewContext.AttributedToLink = attributedTo
		if actorName, foundActorName := actorNames[attributedTo]; foundActorName {
			viewContext.AttributedTo = actorName
		}

		viewContext.Focus = viewObjectRowID == objectRowID

		if objectURL, hasObjectURL := storage.JSONString(objectPayload, "url"); (viewObjectRowID != objectRowID) && hasObjectURL {
			viewContext.LinkView = true
			viewContext.LinkViewURL = objectURL
		}

		viewContext.LinkReply = true
		if viewContext.LinkReply {
			viewContext.LinkReplyURL = fmt.Sprintf("%s?inReplyTo=%s", h.url("compose"), url.QueryEscape(objectID))
		}
		viewContext.LinkAnnounce = true

		if attributedTo == localUserActor.ActorID {
			viewContext.LinkDelete = true
		}

		if published, hasPublished := storage.JSONString(objectPayload, "published"); hasPublished {
			var publishedAt time.Time
			publishedAt, err := time.ParseInLocation("2006-01-02T15:04:05Z", published, time.UTC)
			if err == nil {
				viewContext.HasPublishedAt = true
				viewContext.PublishedAt = publishedAt
			}
		}

		if inReplyTo, hasInReplyTo := storage.JSONString(objectPayload, "inReplyTo"); hasInReplyTo {
			viewContext.HasParent = true
			viewContext.ParentObjectLink = inReplyTo

			if parentObjectRowID, hasParentObjectRowID := objectIDToObjectRowID[inReplyTo]; hasParentObjectRowID {
				if parentRef, hasParentRef := refs[parentObjectRowID]; hasParentRef {

					inReplyToAuthor, hasInReplyToAuthor := storage.JSONString(parentRef, "attributedTo")
					if hasInReplyToAuthor {
						viewContext.ParentAttributedToLink = inReplyToAuthor
						viewContext.ParentAttributedTo = inReplyToAuthor

						if actorName, foundActorName := actorNames[attributedTo]; foundActorName {
							viewContext.ParentAttributedTo = actorName
						}

					}
				}
			}
		}

		objectTag, ok := storage.JSONMapList(objectPayload, "tag")
		if ok {
			for _, tag := range objectTag {
				tagType, hasType := storage.JSONString(tag, "type")
				href, hasHref := storage.JSONString(tag, "href")
				name, hasName := storage.JSONString(tag, "name")
				if hasType && tagType == "Hashtag" && hasName {
					viewContext.Tags = append(viewContext.Tags, name)
				}
				if hasType && tagType == "Mention" && hasHref {
					viewContext.Mentions = append(viewContext.Mentions, href)
				}
			}
		}

		attachments, ok := storage.JSONMapList(objectPayload, "attachment")
		if ok {
			for _, attachment := range attachments {
				mediaURL, hasMediaURL := storage.JSONString(attachment, "url")
				if hasMediaURL && strings.HasPrefix(mediaURL, "https://") {
					blurHash, hasBlurHash := storage.JSONString(attachment, "blurhash")
					mediaType, _ := storage.JSONString(attachment, "mediaType")
					switch mediaType {
					case "image/png", "image/jpeg":
						if strings.HasPrefix(mediaURL, mediaPrefix) {
							viewContext.Media = append(viewContext.Media, viewContentAttachment{
								URL:       mediaURL,
								Thumbnail: fmt.Sprintf("https://%s/asset/blurhash/%s", h.domain, hex.EncodeToString([]byte(blurHash))),
							})
						} else if hasBlurHash {
							viewContext.Media = append(viewContext.Media, viewContentAttachment{
								URL:       mediaURL,
								Thumbnail: fmt.Sprintf("https://%s/asset/blurhash/%s", h.domain, hex.EncodeToString([]byte(blurHash))),
							})
						} else {
							viewContext.Media = append(viewContext.Media, viewContentAttachment{
								URL:       mediaURL,
								Thumbnail: mediaURL,
							})
						}
					}
				}
			}
		}
		if count, hasCount := mappedUserBoosts[viewObjectRowID.String()]; hasCount && count > 0 {
			viewContext.ViewerAnnounced = true
		}
		if count, hasCount := mappedBoosts[viewObjectRowID.String()]; hasCount {
			viewContext.TotalAnnounces = count
		}
		if count, hasCount := mappedReplies[viewObjectRowID.String()]; hasCount {
			viewContext.TotalReplies = count
		}

		return viewContext
	})

	data["root"] = root

	c.HTML(http.StatusOK, "object", data)
}
