package web

import (
	"fmt"
	"net/http"
	"strings"

	"github.com/gin-gonic/gin"
	"go.uber.org/zap"

	"github.com/ngerakines/tavern/errors"
	"github.com/ngerakines/tavern/g"
	"github.com/ngerakines/tavern/storage"
)

func (h handler) nodeInfo(c *gin.Context) {
	c.JSON(http.StatusOK, map[string]interface{}{
		"links": []interface{}{
			map[string]interface{}{
				"rel":  "http://nodeinfo.diaspora.software/ns/schema/2.0",
				"href": fmt.Sprintf("https://%s/nodeinfo/2.0", h.domain),
			},
		},
	})
}

func (h handler) nodeInfoDetails(c *gin.Context) {
	users, err := h.storage.CountUsers(c.Request.Context())
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}
	users30Days, err := h.storage.CountUsersLastMonth(c.Request.Context())
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}
	users183Days, err := h.storage.CountUsersLastHalfYear(c.Request.Context())
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	activities, err := h.storage.CountObjectEvents(c.Request.Context())
	if err != nil {
		h.internalServerErrorJSON(c, err)
		return
	}

	c.JSON(http.StatusOK, map[string]interface{}{
		"version": "2.0",
		"software": map[string]interface{}{
			"name":    "tavern",
			"version": g.Version(),
		},
		"protocols": []string{"activitypub"},
		"usage": map[string]interface{}{
			"users": map[string]interface{}{
				"total":          users,
				"activeMonth":    users30Days,
				"activeHalfyear": users183Days,
			},
			"localPosts": activities,
		},
		"openRegistrations": false,
	})
}

func (h handler) webFinger(c *gin.Context) {
	username, domain, err := fingerUserDomain(c.Query("resource"), h.domain)
	if err != nil {
		h.logger.Error("error parsing resource", zap.Error(err))
		h.writeJRD(c, http.StatusNotFound, nil)
		return
	}

	h.logger.Info("webfinger request", zap.String("resource", c.Query("resource")))

	user, err := h.storage.GetUserByName(c.Request.Context(), username)
	if err != nil {
		if errors.Is(err, errors.NotFoundError{}) {
			h.notFoundJSON(c, nil)
			return
		}
		h.internalServerErrorJSON(c, err)
		return
	}

	response := storage.EmptyPayload()
	response["subject"] = fmt.Sprintf("acct:%s@%s", user.Name, domain)
	response["aliases"] = []string{
		fmt.Sprintf("https://%s/users/%s", domain, user.Name),
	}

	self := storage.EmptyPayload()
	self["rel"] = "self"
	self["type"] = `application/activity+json`
	self["href"] = fmt.Sprintf("https://%s/users/%s", domain, user.Name)
	response["links"] = []storage.Payload{self}

	icon := storage.EmptyPayload()
	icon["type"] = "Image"
	icon["mediaType"] = "image/png"
	icon["url"] = fmt.Sprintf("https://%s/avatar/png/%s", h.domain, user.Name)

	self["icon"] = icon

	h.writeJRD(c, http.StatusOK, response)
}

func fingerUserDomain(input, domain string) (string, string, error) {
	input = strings.TrimPrefix(input, "acct:")

	domainPrefix := fmt.Sprintf("https://%s/users/", domain)
	if strings.HasPrefix(input, domainPrefix) {
		user := strings.TrimPrefix(input, domainPrefix)
		return user, domain, nil
	}

	parts := strings.FieldsFunc(input, func(r rune) bool {
		return r == '@'
	})
	if len(parts) != 2 {
		return "", "", errors.New("malformed resource parameter")
	}
	if parts[1] != domain {
		return "", "", errors.New("malformed resource parameter")
	}
	user := strings.TrimSpace(parts[0])
	if len(user) == 0 {
		return "", "", errors.New("malformed resource parameter")
	}
	return user, domain, nil
}
