package web

import (
	"net/http"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
)

type metricsMiddleware struct {
	requests     *prometheus.CounterVec
	totalTime    *prometheus.HistogramVec
	requestSize  *prometheus.SummaryVec
	responseSize *prometheus.SummaryVec
}

func newMetricsMiddleware(namespace, subsystem string, metricFactory promauto.Factory) metricsMiddleware {
	return metricsMiddleware{
		requests: metricFactory.NewCounterVec(
			prometheus.CounterOpts{
				Namespace: namespace,
				Subsystem: subsystem,
				Name:      "request_total",
				Help:      "Total number of HTTP requests made.",
			}, []string{"status", "endpoint", "method"},
		),
		totalTime: metricFactory.NewHistogramVec(
			prometheus.HistogramOpts{
				Namespace: namespace,
				Subsystem: subsystem,
				Name:      "request_duration",
				Help:      "HTTP request latencies in seconds.",
			}, []string{"status", "endpoint", "method"},
		),
		requestSize: metricFactory.NewSummaryVec(
			prometheus.SummaryOpts{
				Namespace: namespace,
				Subsystem: subsystem,
				Name:      "request_size",
				Help:      "HTTP request sizes in bytes.",
			}, []string{"status", "endpoint", "method"},
		),
		responseSize: metricFactory.NewSummaryVec(
			prometheus.SummaryOpts{
				Namespace: namespace,
				Subsystem: subsystem,
				Name:      "response_size",
				Help:      "HTTP request sizes in bytes.",
			}, []string{"status", "endpoint", "method"},
		),
	}
}

// calcRequestSize returns the size of request object.
func calcRequestSize(r *http.Request) float64 {
	size := 0
	if r.URL != nil {
		size = len(r.URL.String())
	}

	size += len(r.Method)
	size += len(r.Proto)

	for name, values := range r.Header {
		size += len(name)
		for _, value := range values {
			size += len(value)
		}
	}
	size += len(r.Host)

	// r.Form and r.MultipartForm are assumed to be included in r.URL.
	if r.ContentLength != -1 {
		size += int(r.ContentLength)
	}
	return float64(size)
}
